package controller_test

import (
	"testing"

	e "galaxy/error"

	"galaxy/game/internal/controller"
	"galaxy/game/internal/model/game"
	g "galaxy/game/internal/model/game"

	"github.com/google/uuid"
	"github.com/stretchr/testify/assert"
)

func TestGroupUpgradeCost(t *testing.T) {
	sg := &g.ShipGroup{
		Tech: map[g.Tech]g.Float{
			g.TechDrive:   1.0,
			g.TechWeapons: 1.0,
			g.TechShields: 1.0,
			g.TechCargo:   1.0,
		},
		Number: 1,
	}
	assert.Equal(t, 225.0, controller.GroupUpgradeCost(sg, Cruiser, 2.0, 2.0, 2.0, 2.0).UpgradeCost(1))
}

func TestUpgradeMaxShips(t *testing.T) {
	sg := &g.ShipGroup{
		Tech: map[g.Tech]g.Float{
			g.TechDrive:   1.0,
			g.TechWeapons: 1.0,
			g.TechShields: 1.0,
			g.TechCargo:   1.0,
		},
		Number: 10,
	}
	uc := controller.GroupUpgradeCost(sg, Cruiser, 2.0, 2.0, 2.0, 2.0)
	assert.Equal(t, uint(4), uc.UpgradeMaxShips(1000))
}

func TestCurrentUpgradingLevel(t *testing.T) {
	sg := &g.ShipGroup{
		StateUpgrade: nil,
	}
	assert.Equal(t, 0.0, controller.CurrentUpgradingLevel(sg, g.TechDrive))
	assert.Equal(t, 0.0, controller.CurrentUpgradingLevel(sg, g.TechWeapons))
	assert.Equal(t, 0.0, controller.CurrentUpgradingLevel(sg, g.TechShields))
	assert.Equal(t, 0.0, controller.CurrentUpgradingLevel(sg, g.TechCargo))

	sg.StateUpgrade = &g.InUpgrade{
		UpgradeTech: []g.UpgradePreference{
			{Tech: g.TechDrive, Level: 1.5, Cost: 100.1},
		},
	}
	assert.Equal(t, 1.5, controller.CurrentUpgradingLevel(sg, g.TechDrive))
	assert.Equal(t, 0.0, controller.CurrentUpgradingLevel(sg, g.TechWeapons))
	assert.Equal(t, 0.0, controller.CurrentUpgradingLevel(sg, g.TechShields))
	assert.Equal(t, 0.0, controller.CurrentUpgradingLevel(sg, g.TechCargo))

	sg.StateUpgrade.UpgradeTech = append(sg.StateUpgrade.UpgradeTech, g.UpgradePreference{Tech: g.TechCargo, Level: 2.2, Cost: 200.2})
	assert.Equal(t, 1.5, controller.CurrentUpgradingLevel(sg, g.TechDrive))
	assert.Equal(t, 0.0, controller.CurrentUpgradingLevel(sg, g.TechWeapons))
	assert.Equal(t, 0.0, controller.CurrentUpgradingLevel(sg, g.TechShields))
	assert.Equal(t, 2.2, controller.CurrentUpgradingLevel(sg, g.TechCargo))
}

func TestFutureUpgradeLevel(t *testing.T) {
	assert.Equal(t, 0.0, controller.FutureUpgradeLevel(2.0, 2.0, 2.0))
	assert.Equal(t, 0.0, controller.FutureUpgradeLevel(2.0, 2.0, 3.0))
	assert.Equal(t, 1.5, controller.FutureUpgradeLevel(1.5, 2.0, 3.0))
	assert.Equal(t, 2.0, controller.FutureUpgradeLevel(2.5, 1.0, 2.0))
	assert.Equal(t, 2.5, controller.FutureUpgradeLevel(2.5, 1.0, 0.0))
}

func TestUpgradeGroupPreference(t *testing.T) {
	sg := g.ShipGroup{
		Number: 4,
		Tech: g.TechSet{
			g.TechDrive:   1.0,
			g.TechWeapons: 1.0,
			g.TechShields: 1.0,
			g.TechCargo:   1.0,
		},
	}
	assert.Nil(t, sg.StateUpgrade)
	sg = controller.UpgradeGroupPreference(sg, Cruiser, g.TechDrive, 0)
	assert.Nil(t, sg.StateUpgrade)

	sg = controller.UpgradeGroupPreference(sg, Cruiser, g.TechDrive, 2.0)
	assert.NotNil(t, sg.StateUpgrade)
	assert.Equal(t, 300., sg.StateUpgrade.TechCost(g.TechDrive))
	assert.Equal(t, 300., sg.StateUpgrade.Cost())

	sg = controller.UpgradeGroupPreference(sg, Cruiser, g.TechWeapons, 2.0)
	assert.NotNil(t, sg.StateUpgrade)
	assert.Equal(t, 300., sg.StateUpgrade.TechCost(g.TechWeapons))
	assert.Equal(t, 600., sg.StateUpgrade.Cost())

	sg = controller.UpgradeGroupPreference(sg, Cruiser, g.TechShields, 2.0)
	assert.NotNil(t, sg.StateUpgrade)
	assert.Equal(t, 300., sg.StateUpgrade.TechCost(g.TechShields))
	assert.Equal(t, 900., sg.StateUpgrade.Cost())

	sg = controller.UpgradeGroupPreference(sg, Cruiser, g.TechCargo, 2.0)
	assert.NotNil(t, sg.StateUpgrade)
	assert.Equal(t, 0., sg.StateUpgrade.TechCost(g.TechCargo))
	assert.Equal(t, 900., sg.StateUpgrade.Cost())
}

func TestShipGroupUpgrade(t *testing.T) {
	c, g := newCache()
	// group #1 - in_orbit, free to upgrade
	assert.NoError(t, c.CreateShips(Race_0_idx, ShipType_Cruiser, R0_Planet_0_num, 10))
	// group #2 - in_space
	assert.NoError(t, c.CreateShips(Race_0_idx, ShipType_Cruiser, R0_Planet_0_num, 1))
	c.ShipGroup(1).StateInSpace = &InSpace
	// group #3 - in_orbit, foreign planet
	assert.NoError(t, c.CreateShips(Race_0_idx, ShipType_Cruiser, R0_Planet_0_num, 1))
	c.ShipGroup(2).Destination = R1_Planet_1_num

	assert.ErrorContains(t,
		g.ShipGroupUpgrade(UnknownRace, c.ShipGroup(0).ID, "DRIVE", 0),
		e.GenericErrorText(e.ErrInputUnknownRace))
	assert.ErrorContains(t,
		g.ShipGroupUpgrade(Race_Extinct.Name, c.ShipGroup(0).ID, "DRIVE", 0),
		e.GenericErrorText(e.ErrRaceExinct))
	assert.ErrorContains(t,
		g.ShipGroupUpgrade(Race_0.Name, uuid.New(), "DRIVE", 0),
		e.GenericErrorText(e.ErrInputEntityNotExists))
	assert.ErrorContains(t,
		g.ShipGroupUpgrade(Race_0.Name, c.ShipGroup(1).ID, "DRIVE", 0),
		e.GenericErrorText(e.ErrShipsBusy))
	assert.ErrorContains(t,
		g.ShipGroupUpgrade(Race_0.Name, c.ShipGroup(2).ID, "DRIVE", 0),
		e.GenericErrorText(e.ErrInputEntityNotOwned))
	assert.ErrorContains(t,
		g.ShipGroupUpgrade(Race_0.Name, c.ShipGroup(0).ID, "GUN", 0),
		e.GenericErrorText(e.ErrInputTechUnknown))
	assert.ErrorContains(t,
		g.ShipGroupUpgrade(Race_0.Name, c.ShipGroup(0).ID, "CARGO", 0),
		e.GenericErrorText(e.ErrInputUpgradeShipTechNotUsed))
	assert.ErrorContains(t,
		g.ShipGroupUpgrade(Race_0.Name, c.ShipGroup(0).ID, "ALL", 2.0),
		e.GenericErrorText(e.ErrInputUpgradeParameterNotAllowed))
	assert.ErrorContains(t,
		g.ShipGroupUpgrade(Race_0.Name, c.ShipGroup(0).ID, "DRIVE", 2.0),
		e.GenericErrorText(e.ErrInputUpgradeTechLevelInsufficient))
	assert.ErrorContains(t,
		g.ShipGroupUpgrade(Race_0.Name, c.ShipGroup(0).ID, "DRIVE", 1.1),
		e.GenericErrorText(e.ErrInputUpgradeShipsAlreadyUpToDate))

	c.RaceTechLevel(Race_0_idx, game.TechDrive, 10.0)
	assert.Equal(t, 10.0, c.Race(Race_0_idx).TechLevel(game.TechDrive))
	assert.ErrorContains(t,
		g.ShipGroupUpgrade(Race_0.Name, c.ShipGroup(0).ID, "DRIVE", 10.0),
		e.GenericErrorText(e.ErrUpgradeInsufficientResources))

	assert.NoError(t, g.ShipGroupUpgrade(Race_0.Name, c.ShipGroup(0).ID, "DRIVE", 1.3))
	assert.Equal(t, game.StateInOrbit, c.ShipGroup(0).State())
	assert.Equal(t, uint(6), c.ShipGroup(0).Number)
	assert.Equal(t, game.StateUpgrade, c.ShipGroup(3).State())
	assert.Equal(t, uint(4), c.ShipGroup(3).Number)
	assert.NotNil(t, c.ShipGroup(3).StateUpgrade)
	assert.Equal(t, 1.3, c.ShipGroup(3).StateUpgrade.UpgradeTech[0].Level.F())
	assert.Equal(t, "DRIVE", c.ShipGroup(3).StateUpgrade.UpgradeTech[0].Tech.String())

	assert.ErrorContains(t,
		g.ShipGroupUpgrade(Race_0.Name, c.ShipGroup(3).ID, "DRIVE", 1.3),
		e.GenericErrorText(e.ErrShipsBusy))
}
